#!/usr/bin/env python3

#
# copyright_plugin.py
#
# Copyright 2020 Christian Hergert <chergert@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import GLib, GObject, Gio
from gi.repository import Gtk
from gi.repository import Ide

import re
import time

_YEAR_REGEX = r1 = re.compile('([0-9]{4})')
_MAX_LINE = 100
_ = Ide.gettext

class CopyrightBufferAddin(Ide.Object, Ide.BufferAddin):
    settings = None

    def do_load(self, buffer):
        self.settings = Gio.Settings.new('org.gnome.builder.plugins.copyright')

    def do_unload(self, buffer):
        self.settings = None

    def do_save_file(self, buffer, file):
        """
        If there is a copyright block within the file and we find the users
        name within it, we might want to update the year to include the
        current year.
        """
        # Stop if we aren't enabled
        if not self.settings.get_boolean('update-on-save'):
            return

        # Stop if we can't discover the user's real name
        name = GLib.get_real_name()
        if name == 'Unknown':
            return

        year = time.strftime('%Y')
        iter = buffer.get_start_iter()
        limit = buffer.get_iter_at_line(_MAX_LINE)

        while True:
            # Stop if we've past our limit
            if iter.compare(limit) >= 0:
                break

            # Stop If we don't find the user's full name
            ret = iter.forward_search(name, Gtk.TextSearchFlags.TEXT_ONLY, limit)
            if not ret:
                break

            # Get the whole line text
            match_begin, match_end = ret
            match_begin.set_line_offset(0)
            if not match_end.ends_line():
                match_end.forward_to_line_end()
            text = match_begin.get_slice(match_end)

            # Split based on 4-digit years
            parts = _YEAR_REGEX.split(text)

            # Ignore if this year is already represented
            if year in parts:
                break

            # If we have at least 2 years, we can update them
            if len(parts) >= 2:
                if '-' in parts:
                    parts[parts.index('-')+1] = year
                else:
                    parts.insert(2, '-')
                    parts.insert(3, year)

                text = ''.join(parts)

                # Replace the line text in a single undo action
                buffer.begin_user_action()
                buffer.delete(match_begin, match_end)
                buffer.insert(match_begin, text)
                buffer.end_user_action()

                break

            # Move to the end of the line
            iter = match_end

class CopyrightPreferencesAddin(GObject.Object, Ide.PreferencesAddin):
    def do_load(self, prefs):
        self.update_on_save = prefs.add_switch(
                "editor", "general",
                "org.gnome.builder.plugins.copyright",
                "update-on-save",
                None,
                "false",
                _("Update Copyright"),
                _("When saving a file Builder will automatically update copyright information for you"),
                # translators: these are keywords used to search for preferences
                _("update copyright save"),
                10)

    def do_unload(self, prefs):
        prefs.remove_id(self.update_on_save)



#!/usr/bin/env python3
#
# vala_pack_plugin.py
#
# Copyright 2020 Christian Hergert <christian@hergert.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import GLib
from gi.repository import Gdk
from gi.repository import Ide

_ERROR_REGEX = (
    "(?<filename>[a-zA-Z0-9\\-\\.\\/_]+.vala):" +
    "(?<line>\\d+).(?<column>\\d+)-(?<line2>\\d+).(?<column2>\\d+): " +
    "(?<level>[\\w\\s]+): " +
    "(?<message>.*)"
)

class ValaPipelineAddin(Ide.Object, Ide.PipelineAddin):
    error_format = 0

    def do_load(self, pipeline):
        self.error_format = pipeline.add_error_format(_ERROR_REGEX, GLib.RegexCompileFlags.OPTIMIZE)

    def do_unload(self, pipeline):
        pipeline.remove_error_format(self.error_format)

def copy_indent(iter):
    begin = iter.copy()
    begin.set_line_offset(0)
    end = begin.copy()
    while not end.ends_line() and end.get_char().isspace() and end.forward_char():
        # Do nothing
        pass
    return begin.get_slice(end)

def get_line_text(iter):
    begin = iter.copy()
    end = iter.copy()
    begin.set_line_offset(0)
    if not end.ends_line():
        end.forward_to_line_end()
    return begin.get_slice(end)

def is_newline_in_braces(iter):
    prev = iter.copy()
    next = iter.copy()
    prev.backward_char()
    next.forward_char()
    return prev.get_char() == '{' and iter.get_char() == '\n' and next.get_char() == '}'

def is_newline_keyval(keyval):
    return keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter)

def in_comment(iter):
    buffer = iter.get_buffer()
    copy = iter.copy()
    copy.backward_char()
    return buffer.iter_has_context_class(copy, 'comment')

def indent_comment(iter):
    line = get_line_text(iter).strip()

    # Continue with another single line comment
    if line.startswith('//'):
        return copy_indent(iter) + '// '

    # Comment is closed, copy indent, possibly trimming extra space
    if line.endswith('*/'):
        if line.startswith('*'):
            str = copy_indent(iter)
            if str.endswith(' '):
                str = str[:-1]
            return str

    if line.startswith('/*') and not line.endswith('*/'):
        return copy_indent(iter) + ' * '
    elif line.startswith('*'):
        return copy_indent(iter) + '* '

    return copy_indent(iter)

class ValaIndenter(Ide.Object, Ide.Indenter):

    def do_is_trigger(self, key):
        return key.keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter, Gdk.KEY_slash)

    def do_format(self, text_view, begin, end, key):
        was_newline = is_newline_keyval(key.keyval)
        copy = end.copy()

        # Move us back to the jsut inserted character
        copy.backward_char()

        # If we are in a comment, continue the indentation
        if in_comment(copy):
            # Maybe close a multi-line comment
            if copy.get_char() == '/':
                close = copy.copy()

                if close.backward_char() and close.get_char() == ' ' and close.backward_char() and close.get_char() == '*' :
                    begin.backward_char()
                    begin.backward_char()
                    return ('/', 0)

            if was_newline:
                return (indent_comment(copy), 0)

        if is_newline_in_braces(copy):
            prefix = copy_indent(copy)

            if text_view.get_insert_spaces_instead_of_tabs():
                indent = "    "
            else:
                indent = '\t'

            return (prefix + indent + "\n" + prefix, -(len(prefix) + 1))

        if was_newline:
            return (copy_indent(copy), 0)

        return (None, 0)

